/*
  This is globals.h
  
  Coxeter version 1.0  Copyright (C) 1991 Fokko du Cloux
  See file main.c for full copyright notice
*/

#define BLOCKSIZE 4096
#define LINESIZE 80        /* buffers for i/o */
#define LLINESIZE 1024     /* for long lines */

#define CALLOC(_p,_n,_x,statement) if ((_p = (_x *)calloc((unsigned)_n,sizeof(_x)))==0) printf("\nout of memory\n\n"),exit();

#define FREE(p,n,p_t) free(p);

#define CHARBITS 8                        /* number of bits in a char */
#define BITS(x_t) CHARBITS*sizeof(x_t)    /* size of x_t in bits */

#define VERSION "1.0"

#define small_size 1
#define medium_size 2
#define large_size 3
#define huge_size 4

/*************************************************************************

                        global variables

*************************************************************************/

extern typeid_t type;
extern rankid_t rank;
extern status_t status;
extern int representation;
extern int dump_flag;

extern weyldata_t **head;
extern state_t **table;
extern weylsize_t wsize;
extern weylarr_t longarray;
extern lengthcount_t maxlength;
extern weylsize_t *wsubsize;
extern weylsize_t *betti;
extern weylsize_t *firstlength;
extern weylword_t print_string;
extern sflags_t *star;
extern bettipol_t *qbetti;

extern weylarr_t lexorder;
extern weylnbr_t *natorder;
extern lflags_t **brorder;
extern unsigned long brsize;
extern unsigned short *rowlength;
extern downflags_t *downflags;
extern rowhead_t *prbruhat;
extern unsigned char *charmask;
extern downflags_t *dfmask;
extern downflags_t rightmask;
extern downflags_t leftmask;
extern lflags_t *wordmask;
extern weylnbr_t *inverse;
extern weylnbr_t *lexprod;
extern int wordsize;

extern unsigned long wrowlength;
extern rowhead_t *kl_list;
extern lflags_t **dflist;
extern unsigned long elsize;
extern klcell_t *klstore;
extern zoneinfo_t *klzone;	
extern bettipol_t *bettipol;
extern betticell_t *bettistore;
extern zoneinfo_t *bettizone;


/* functions from commands.c */

int Run();
int Help();

/* functions from interface.c */

int GetInfo();
int InitInterface();
int GetType();
int GetRank();
int CheckChoice();
int PrintString();
int SprintString();
int PrintArray();
int GetInput();
int GetPol();
int GetCell();
int GetMultiplicities();
int PrintMultiplicities();
int GetAll();
int PrintCell();
int TeXCell();
int TestPositive();
int PrintTree();
int PrintPol();
int PrintBPol();
int SprintPol();
int CompletePresentation();
int MathPrint();

/* functions from Weyl.c */

int Weyl();
int MakeAutomaton();
int ArrayToString();
int NbrToStringpiece();
int Shift();
int StringShift();
int StringToArray();
int RStringToArray();
int Prod();
int Inverse();
int Mult();
int StringMult();
int NormalForm();
generator_t Firstterm();
generator_t Lastterm();
int Longest();
int NaturalOrder();
int MakeBetti();
int InitAutomaton();
int DestroyAutomaton();
int GetInitData();
typeid_t GetSubType();
quotsize_t GetQuotSize();
lengthcount_t GetMaxql();
int BettiData();
int Cox();
typeid_t CompType();
int BPolMult();
int BPolDiv();
int BettiPol();
int BitCount();
int WeylnbrToArray();
int ArrayToWeylnbr();

/* functions from Bruhat.c */

int Bruhat();
int PreBruhat();
int MakeBrOrder();
int MakeBrPiece();
int InitBruhat();
int FreeBruhat();
static InitConstants();
int MakeNLOrders();
weylnbr_t LexOrder();
int MakeDownFlags();
int MakeInverse();
int MakePrBruhat();
struct CutData CutOff();
int LexRProd();
int LexLProd();
int MakeLexProd();

/* functions from kl.c */

int KL();
static InitConstants();
int InitKL();
int MakeExtrList();
int MakeExtrLists_new();
int Extremals();
rowsize_t RowBits();
int MakeKLLists();
int MakeMuList();
int MakeKLList();
klpol_t *KLpol();
weylnbr_t Extr();
klpol_t *GetAdress();
int Compare();
klcell_t *MakeSon();
int BettiNumbers();
static Add();
static Zero();
int MakeBCells();
bettipol_t GetBAdress();
int BCompare();
betticell_t *MakeBSon();
sflags_t RtoL();
int BKLPolMult();
int DumpFiles();
int ReadFiles();
klcoeff_t Multiplicity();
int WriteTree();

/* functions from memory.c */

zoneinfo_t *AllocZone();
int NewBlock();
int FreeZone();
