/*
  This is interface.c
  
  Coxeter version 1.0  Copyright (C) 1991 Fokko du Cloux
  See file main.c for full copyright notice
*/

#include <stdio.h>
#include "interface.h"

/*
  This module contains the routines for input and output, to be called by
  commands.h. We have tried to use routines from stdio.h only, in order
  to ensure maximal portability.

  Output to files is usually in TeX-format -- the files should go right
  thru the TeX-compiler. Some binary files are created by routines in
  kl.c, either by default or on request. This should probably be improved,
  when a consensus on the format for such files has been reached. Right
  now, the intention was mostly to avoid recomputing them over and over
  again for large groups.
*/

InitInterface()

/*
  Does some allocations and initializations. Can be called only after type
  and rank are known.
*/

{

  CALLOC(input_array,rank,weylpiece_t,;)
  CALLOC(itoa_array,rank,weylpiece_t,;)

}    /* end of InitInterface */

GetType(c)

/**********************************************************************

	Gets "type" from the user's terminal. Filters according to the
present state of the program.

***********************************************************************/

     typeid_t *c;

{
static char buf[LINESIZE];

printf("\nType : ");
gets(buf);
*c = buf[0];

while ((*c<'A')|(*c>'H'))
	{
	printf("\nThe types allowed for now are A,B,C,D,E,F,G,H \
(uppercase please) : ");
	gets(buf);
	*c = buf[0];
	}

return(0);

}	/* end of GetType */


GetRank(l)

/**********************************************************************

	Gets "rank" from the user's terminal. Filters according to the
present state of the program.

************************************************************************/

	rankid_t *l;

{

static char buf[LINESIZE];

printf("\nRank : ");
gets(buf);
*l=atoi(buf);

switch (type)
	{
	case 'A':
		while ((*l<1)|(*l>255))
			{
			printf("\nIn type A the rank should be between 1 and 255 ... : ");
			gets(buf);
			*l=atoi(buf);
			}
		return(0);
	
	case 'B':
	case 'C':
		while ((*l<1)|(*l>255))
			{
			printf("\nIn type B the rank should be between 1 and 255 ... : ");
			gets(buf);
			*l=atoi(buf);
			}
		if (*l==1)
			{
			type = 'A';
			}
		return(0);
		
	case 'D':
		while ((*l<2)|(*l>255))
			{
			printf("\nIn type D the rank should be between 2 and 255 ... : ");
			gets(buf);
			*l=atoi(buf);
			}
		return(0);
		
	case 'E':
		while ((*l<3)|(*l>8))
			{
			printf("\nIn type E the rank should be between 3 and 8 ... : ");
			gets(buf);
			*l=atoi(buf);
			}
		return(0);
		
	case 'F':
		while ((*l<3)|(*l>4))
			{
			printf("\nIn type F the rank should be 3 or 4 ... : ");
			gets(buf);
			*l=atoi(buf);
			}
		return(0);
		
	case 'G':
		if (*l!=2)
			{
			printf("\nIn type G the rank should be 2, so I'm changing\n\
the rank to 2");
			*l = 2;
			}
		return(0);
		
	case 'H':
		while ((*l<2)|(*l>4))
			{
			printf("\nIn type H the rank should be 2,3 or 4 ... : ");
			gets(buf);
			*l=atoi(buf);
			}
		return(0);
	}

}	/* end of GetRank */


GetRepresentation(r)

/**********************************************************************

	Gets from the user the way he wants to enter the generators.

***********************************************************************/

	int *r;

{
static char buf[LINESIZE];
int j;

printf("\n\n");

printf("In type %c%d the generators are numbered as follows :\n\n",type,rank);

switch (type)
	{
	
	case 'A':
		printf("\t\t");
		if (rank<=8)
			{
			printf("1");
			for (j=2;j<=rank;j++)
				{
				printf(" - %d",j);
				}
			}
		else
			{
			printf("1 - 2 - ... - %d",rank);
			}
		printf("\n\n");
		break;
		
	case 'B':
	case 'C':
		printf("\t\t");
		if (rank<=8)
			{
			printf("1 = 2");
			for (j=3;j<=rank;j++)
				{
				printf(" - %d",j);
				}
			}
		else
			{
			printf("1 = 2 - ... - %d",rank);
			}
		printf("\n\n(the double edge is the one marked '4' in the \
Coxeter graph)\n");
		break;
		
	case 'D':
		printf("\t\t");
		if (rank == 2)
			{
			printf("1   2\n\n");
			}
		else if (rank<=8)
			{
			printf("1 - 3");
			for (j=4;j<=rank;j++)
				{
				printf(" - %d",j);
				}
			printf("\n\t\t    |");
			printf("\n\t\t    2\n\n");
			}
		else
			{
			printf("1 - 3 - ... - %d",rank);
			printf("\n\t\t    |");
			printf("\n\t\t    2\n\n");
			}
		break;
		
	case 'E':
		printf("\t\t");
			if (rank == 3)
				{
				printf("1 - 3  2\n\n");
				}
			else
				{
				printf("1 - 3 - 4");
				for (j=5;j<=rank;j++)
					{
					printf(" - %d",j);
					}
				printf("\n\t\t        |");
				printf("\n\t\t        2\n\n");
				}
			break;
		
	case 'F':
		if (rank == 3)
			{
			printf("\t\t1 - 2 = 3");
			}
		else
			{
			printf("\t\t1 - 2 = 3 - 4");
			}
		printf("\n\n(the double edge is the one marked '4' in the \
Coxeter graph)\n");
		break;
		
	case 'G':
		printf("\t\t1 = 2");
		printf("\n\n(the double edge is marked '6' in the \
Coxeter graph)\n");
		break;
		
	case 'H':
		printf("\t\t");
		printf("1 = 2");
		for (j=3;j<=rank;j++)
			{
			printf(" - %d",j);
			}
		printf("\n\n(the double edge is the one marked '5' in the \
Coxeter graph)\n");
		break;
	}

printf("(this cannot be changed for the time being)\n\n\
You have two choices for the way to represent the generators :\n\n\
\t1. One-digit hex (1 to f inclusive)\n\n\
\t2. Lower-case letters (a to z inclusive)\n\n\
Your choice may be modified with the rep command\n\
Choose the way you prefer : ");

gets(buf);
*r = atoi(buf);
while ((*r<1)|(*r>2))
	{
	printf("\nPlease type 1 or 2 : ");
	gets(buf);
	*r = atoi(buf);
	}
return(0);

}	/* end of GetRepresentation */


CheckChoice(type,rank,maxsize)

/*
  Checks if the size of the group is larger than maxsize. If we stick to
  the standard automata, this is the only possible reason for overflow.
  Should be checked once after the type and rank are chosen, with
  maxsize the largest weylsize_t, and once if kl-computations are
  undertaken, with maxsize the largest weylnbr_t. The return-value is 0
  if no overflow, the maximal allowable rank plus one if there is
  overflow.
*/

     typeid_t type;
     rankid_t rank;
     unsigned long maxsize;

{
  unsigned long size;
  rankid_t l;

  /* put in maxsize the largest unsigned quantity it will hold */
  
  maxsize = 0;
  maxsize = ~maxsize;  

/*
  Check if weylsize_t will hold wsize. Also, check if weylpiece_t will hold
  enough elements.
*/

switch (type)
  {
  case 'A':
    size = 1;
    for (l=1; l <= rank; l++)
      {
	if (size > maxsize/(l+1)) /* overflow */
	  {
	    return(l);
	  }
	size *= l+1;
      }
    break;
  case 'B':
  case 'C':
    size = 1;
    for (l=1; l <= rank; l++)
      {
	if (size > maxsize/(2*l))  /* overflow */
	  {
	    return(l);
	  }
	size *= 2*l;
      }
    break;
  case 'D':
    size = 1;
    for (l=2; l <= rank; l++)
      {
	if (size > maxsize/(2*l))  /* overflow */
	  {
	    return(l);
	  }
	size *= 2*l;
      }
    break;
  case 'E':
    {
      static unsigned q[5] = {10,16,27,56,240};
      
      size = 12;  /* size of E3 */

      for (l=4; l <= rank; l++)
	{
	  if (size >= maxsize/q[l-4])  /* overflow */
	    {
	      return(l);
	    }
	  size *= q[l-4];
	}
    }
    break;
  case 'F':  /* in the remaining cases, overflow is impossible */
  case 'G':
  case 'H':
    break;
  }

  return(0);
  
}	/* end of CheckChoice */


GetInput(a)

/* 
  Reads a Weyl group element from the standard input, according to the current
  representation. It keeps prompting until the user types a valid string. A
  certain number of shortcuts are supported (see the help message for compute).
  Replaces both normalize and Multiply from earlier versions.
*/

    weylarr_t a;

{
  char buf[LLINESIZE];
  int scan;
  
  GetLine(buf,LLINESIZE,stdin);
  while (scan = ScanLine(buf))
    {
      switch (scan)
	{
	case 1:
	  printf("bad input -- try again :\n");
	  break;
	case 2:
	  printf("number too big -- should not excede wsize-1 = %ld\n",wsize-1);
	  break;
	case 3:
	  printf("Error -- EOF before newline was found\n");
	  return(0);
	}
      GetLine(buf,LLINESIZE,stdin);
    }

  InputToArray(buf,a);

}    /* end of GetInput */


static GetLine(s,n,f)

/*
  This is essentially a customized version of fgets; the main point is that
  we want to get rid of the newline in all cases; but we also want the
  protection from overflow provided by fgets. In practice, the maximum
  size of the input may be determined by the input buffer for stdin. Returns
  the number of characters read.
*/

     char *s;
     unsigned n;
     FILE *f;

{

  unsigned j;
  int c;

  j = 0;

  do
    {
      if ((c = getc(f))==EOF)  /* only happens if file ends without a newline */
	{
	  return(3);
	}
      *s = c;
      j++;
    }
  while ((*(s++)!='\n')&&(j < n-1));

  if (*(--s)=='\n')
    {
      j--;
      *s = '\0';
    }
  else
    {
      *(++s) = '\0';
    }

  return(j);

}    /* end of GetLine */


static ScanLine(line)

/*
  Scans the input taken by GetLine for correctness.  It checks the range
  of the characters (according to the chosen representation and current
  rank), and the range of the numbers following # and ^ .  In both
  cases, we note that the maximum size is wsize (in the case of ^, we
  may always reduce modulo wsize -- this could be done automatically,
  but it didn't seem worthwile).

  In the process of examining the input, the string s is rewritten :
  the generators are transformed into characters from 1 to rank, and
  white space is deleted.
*/

     char * line;

{
  char *t,*s;
  char c1;
  static casewarning = 0;
  int overflow;
  weylnbr_t w;
  int d;

  t = line;
  s = line;

  while (isspace(*s))  /* remove leading white space */
    {
      s++;
    }

  for(;*s;s++)
    {
      if (isspace(*s))	/* transform into ordinary space */
	{
	  *(t++) = ' ';
	}
      else if (*s == '*' || *s=='!')
	{
	  *(t++) = *s;
	}
      else if (*s=='#' || *s=='^') /* read a number */
	{
	  w = 0;
	  *(t++) = *s;
	  overflow = 0;
	  for (s++;isdigit(*s);s++)
	    {
	      if (overflow == 0)
		{
		  d = *s - '0';
		  if(w > (wsize/10) || d > wsize || 10*w >= wsize-d)
		    {
		      overflow = 1;
		    }
		  else
		    {
		      w = 10*w + d;
		      *(t++) = *s;
		    }
		}
	    }
	  s--;    /* back up by one */
	  if (overflow)
	    {
	      *t = '\0';
	      return(2);  /* bad input */
	    }
	}
      else
	{
	  if (isupper(*s))
	    {
	      *s = tolower(*s);
	      if (casewarning == 0)
		{
		  printf("Warning : input was converted to lower case\n");
		  casewarning = 1;
		}
	    }
	  c1 = GetGenerator(*s,status.representation);
	  if (c1)
	    {
	      *(t++) = c1;
	    }
	  else			/* bad input */
	    {
	      *t = '\0';
	      return(1);
	    }
	}
    }

  *t = '\0';

  t = line;
  s = line;

  for (;*s;s++)
    {
      if (*s == ' ')
	{
	  if (*(s+1) > rank) /* spaces are kept only in front of a string*/
	    continue;
	}
      *(t++) = *(s);
    }

  *t = '\0';
  return(0);
  
}    /* end of ScanLine */


static GetGenerator(c,r)

/*
  Checks whether c is a valid generator in the representation r; returns the
  corresponding generator if yes, 0 if no.
*/

     char c;
     int r;

{
  int d;

  switch (r)
    {
    case 1:  /* c must be hex digit */
      if (isxdigit(c))
	{
	  if (isdigit(c))
	    {
	      d = c-'1';
	    }
	  else
	    {
	      d = 10 + c-'a';
	    }
	  if  (d < rank)
	    {
	      return(d+1);
	    }
	  else  /* out of range */
	    {
	      return(0);
	    }
	}
      else  /* not hex digit */
	{
	  return(0);
	}
      break;
    case 2:  /* c must be lower-case letter */
      if (islower(c))
	{
	  d = c-'a';
	  if (d < rank)
	    {
	      return(d+1);
	    }
	  else  /* out of range */
	    {
	      return(0);
	    }
	}
      else  /* not lower case letter */
	{
	  return(0);
	}
      break;
    }

}    /* end of GetGenerator */


static InputToArray(input,a)

/*
  Transforms the output of ScanLine into a weylarr_t. 
*/

     char *input;
     weylarr_t a;

{
  rankid_t l;
  generator_t s;
  weylnbr_t w;
  unsigned long p;
  int d;

  for (l=0;l<rank;l++)  /* clean up a */
    {
      a[l] = 0;
      itoa_array[l] = 0;  /* holds current piece */
    }

  for (;s = *input;input++)
    {
      switch (s)
	{
	case '*':
	  /*
	     note : we do not flush here; in other words, * should be seen
	     as a unary operation whose argument is to the left of it. If
	     nothing is to the left, then it will act as a token. Example :
	     *#321^5 is *(#321^5); #321*^5 is (#321*)^5.
	  */
	  Mult(itoa_array, longarray, itoa_array);
	  break;
        case '!':
          /*  also a unary operator  */
          Inverse(itoa_array,itoa_array);
          break;
	case '#':
	  Flush(a,itoa_array,a);   /* flush itoa_array */
	  for (w=0,input++;isdigit(*input);input++)  /* get w */
	    {
	      d = *input - '0';
	      w = 10*w + d;
	    }
	  input--;  /* need to back up */
	  WeylnbrToArray(w,itoa_array);
	  break;
	case '^':
	  for (p=0,input++;isdigit(*input);input++)  /* get p */
	    {
	      d = *input - '0';
	      p = 10*p + d;
	    }
	  input--;  /* back up by one */
	  Power(itoa_array,p,itoa_array);
	  /* we don't flush here; allows for iterated powers */
	  break;
	case ' ':
	  Flush(a,itoa_array,a);    /* flush itoa_array */
	  break;
	default:
	  Shift(itoa_array,s);
	  break;
	}
    }

  Mult(a,itoa_array,a);

  return(0);
  
}    /* end of InputToArray */


static Flush(a,b,c)

/*
  Flushes array b into a; result goes to c.
*/

     weylarr_t a,b,c;

{

  rankid_t l;

  Mult(a,b,c);

  for (l=0;l<rank;l++)
    {
      b[l] = 0;
    }

}    /* end of Flush */


PrintArray(a,outputfile)

/*
  Prints the array a, according to the current representation. It also
  prints the corresponding number in lexicographical order.
*/

     weylarr_t a;
     FILE *outputfile;

{

  ArrayToString(a,print_string);
  PrintString(outputfile,print_string);

}    /* end of PrintArray */


PrintString(outputfile,f)

/*
  Prints a Weyl group element, represented as a string, according
  to the chosen representation.
*/
	
	FILE *outputfile;
	weylword_t f;

{

switch (representation)
	{
	case 1:
		for (;*f!=0;f++)
			{
			fprintf(outputfile,"%x",*f);
			}
		break;
	case 2:
		for (;*f!=0;f++)
			{
			fprintf(outputfile,"%c",*f+'a'-1);
			}
		break;
	}
}	/* end of PrintString */


SprintString(s,f)

/*********************************************************************

	Like PrintString, but with output to a string. Returns the number
of characters printed, including the null terminator.

**********************************************************************/
	
	char *s;
	weylword_t f;

{
register int c=1;	/* counts the number of characters */

switch (representation)
	{
	case 1:
		for (;*f!=0;f++,c++)
			{
			if (*f<=9)
				{
				*(s++) = '0' + (*f);
				}
			else
				{
				*(s++) = 'a' + (*f) - 10;
				}
			}
		break;
	case 2:
		for (;*f!=0;f++,c++)
			{
			*(s++) = 'a' + (*f) - 1;
			}
		break;
	}

*s = '\0';
return(c);

}	/* end of SprintString */


GetPol()

/**********************************************************************

	Gives "element-by-element" access to the KL file.

***********************************************************************/

{
  int ly,lw;
  weylnbr_t y,w;
  weylarr_t yarray,warray;
  klpol_t pol,*polhandle;
  char buf[LINESIZE];

  CALLOC(yarray,rank,weylpiece_t,;)
  CALLOC(warray,rank,weylpiece_t,;)

  printf("First element (finish with carriage-return) : ");
  GetInput(yarray);

  printf("Second element : ");
  GetInput(warray);

  y = LexOrder(yarray);
  w = LexOrder(warray);

  if (polhandle = KLpol(y,w))
    {
      pol = *polhandle;
      PrintPol(stdout,pol);
      for (ly=0;y>=firstlength[ly+1];ly++) /* find length of y in ly */
	;
      for (lw=ly;w>=firstlength[lw+1];lw++) /* find length of w in lw */
	;
      if (2**pol==(lw-ly-1))
	{
	  printf(" *");
	}
    }
  else
    {
      printf("0 (the two elements are not in bruhat order)");
    }

  printf("\n");
}				/* end of GetPol */


GetCell()

/**********************************************************************

	Gives "cell-by-cell" access to the KL file.

***********************************************************************/

{
  weylarr_t warray;
  weylnbr_t w;
  FILE *outputfile;
  static int screenoutput;
  char filename[LINESIZE];
  char buf[LINESIZE];

  CALLOC(homology,maxlength+2,weylsize_t,;)
  CALLOC(Ihomology,maxlength+2,weylsize_t,;)
  CALLOC(warray,rank,weylpiece_t,;)

  printf("Enter your element (finish with a carriage-return) :\n");
  GetInput(warray);

  if (screenoutput==0)
    {
      printf("Name an output file (hit return for stdout):\n");
      gets(filename);
      if (*filename=='\0')
	{
	  outputfile = stdout;
	  screenoutput = 1;
	}
      else
	{
	  outputfile = fopen(filename,"w");
	}
    }
  else
    {
      outputfile = stdout;
    }
  w = LexOrder(warray);
  if (screenoutput)
    {
      PrintCell(outputfile,w);
    }
  else
    {
      fprintf(outputfile,"\\hsize16truecm\\vsize23truecm");
      /* works for A4 printout */
      TeXCell(outputfile,w);
      fprintf(outputfile,"\n\n\\bye");
      fclose(outputfile);
    }

  printf("\n\n");
	
  free(homology);
  free(Ihomology);

}				/* end of GetCell */


PrintCell(outputfile,w)

/**********************************************************************

	Prints the output for cell no. w in the file outputfile

***********************************************************************/

	FILE *outputfile;
	weylnbr_t w;

{
kldata_t *kld;
weylnbr_t y;
static weylletter_t f[MAXWORD+1];
int p,q,i;
char buf[LINESIZE],*bufpos;
unsigned rowsize,rj;
weylsize_t *h,*ih;
int iflag;

iflag = 0;

p = ArrayToString(lexorder+rank*w,f);
fprintf(outputfile,"\nw = ");
PrintString(outputfile,f);
fprintf(outputfile,"   (element no. %ld in lexorder)\n",(unsigned long)w);

if (w > inverse[w])
  {
    w = inverse[w];
    iflag = 1;
  }

kld = (kldata_t *)kl_list[w].row;
rowsize = kl_list[w].size;

for (rj=0,i=0;rj<rowsize;rj++,kld++,i++)
	{
	  if (iflag)
	    {
	      y = inverse[kld->y];
	    }
	  else
	    {
	      y =kld->y;
	    }
	  bufpos = buf;
	  sprintf(bufpos,"y = ");
	  for (;*bufpos!=0;bufpos++)
	    ;
	  q = ArrayToString(lexorder+rank*(y),f);
	  bufpos += SprintString(bufpos,f)-1;
	  sprintf(bufpos," : ");
	  for (;*bufpos!=0;bufpos++)
	    ;
	  bufpos += SprintPol(bufpos,*(kld->pol))-1;
	  if (2*(*(kld->pol)[0])==p-q-1) /* we have \mu_{y,w} != 0 */
	    {
	      sprintf(bufpos," *");
	      for (;*bufpos!=0;bufpos++)
		;
	    }
	  if ((i%2) == 0)
	    {
	      fprintf(outputfile,"\n");
	    }
	  fprintf(outputfile,"%-39s",buf);
	}
	
BettiNumbers(w,homology,Ihomology);

h = homology+1;
ih = Ihomology+1;

fprintf(outputfile,"\n\nBetti numbers :\n");
for (q=0;q<=p;q++)
	{
	if ((q%5)==0)
		{
		fprintf(outputfile,"\n");
		}
	sprintf(buf,"h[%d] = %ld",q,h[q]);
	fprintf(outputfile,"%-15s",buf);
	}
	
fprintf(outputfile,"\n\nIH Betti numbers :\n");
for (q=0;q<=p;q++)
	{
	if ((q%5)==0)
		{
		fprintf(outputfile,"\n");
		}
	sprintf(buf,"h[%d] = %ld",q,ih[q]);
	fprintf(outputfile,"%-15s",buf);
	}

}	/* end of PrintCell */


TeXCell(outputfile,w)

/**********************************************************************

	Same as PrintCell, but makes a TeX-file.

***********************************************************************/

	FILE *outputfile;
	weylnbr_t w;

{
kldata_t *kld;
weylnbr_t y;
static weylletter_t f[MAXWORD+1];
int p,q,r,i,j;
unsigned rowsize,rj;
downflags_t dflags;
int empty,iflag;
weylsize_t *h,*ih;

iflag = 0;

p = ArrayToString(lexorder+rank*w,f);
fprintf(outputfile,"\n\\noindent$w = ");
PrintString(outputfile,f);
fprintf(outputfile,"$\\quad(");
if (w==0)
	{
	fprintf(outputfile,"element ");
	}
fprintf(outputfile,"no.\\ $%ld$",(unsigned long)w);
if (w==0)
	{
	fprintf(outputfile," in lexorder");
	}
fprintf(outputfile,")");
dflags = downflags[w];
empty = 1;
fprintf(outputfile,"\\quad L : $(");
for (j=rank+1;j<=2*rank;j++)
	{
	if(dflags&dfmask[j])
		{
		if(empty)
			{
			empty = 0;
			fprintf(outputfile,"%d",j-rank);
			}
		else
			{
			fprintf(outputfile,",%d",j-rank);
			}
		}
	}
fprintf(outputfile,")$");
empty = 1;
fprintf(outputfile,"\\quad R : $(");
for (j=1;j<=rank;j++)
	{
	if(dflags&dfmask[j])
		{
		if(empty)
			{
			empty = 0;
			fprintf(outputfile,"%d",j);
			}
		else
			{
			fprintf(outputfile,",%d",j);
			}
		}
	}
fprintf(outputfile,")$\n");

fprintf(outputfile,"\n\\medskip\n");

if (w > inverse[w])
  {
    w = inverse[w];
    iflag = 1;
  }

fprintf(outputfile,"\n\\begingroup\n\
\\settabs 2 \\columns");

kld = (kldata_t *)kl_list[w].row;
rowsize = kl_list[w].size;

for (rj=0,i=0;rj<rowsize;rj++,kld++)
	{
	  if (iflag)
	    {
	      y = inverse[kld->y];
	    }
	  else
	    {
	      y = kld->y;
	    }
	  if (i%2==0)		/* start new line */
	    {
	      fprintf(outputfile,"\n\\+");
	    }
	  else
	    {
	      fprintf(outputfile,"&");
	    }
	  q = ArrayToString(lexorder+rank*(y),f);
	  if (2*(*(kld->pol)[0])==p-q-1) /* \mu_{y,w} != 0 */
	    {
	      fprintf(outputfile,"\\llap{$\\bullet$\\enskip}");
	    }
	  fprintf(outputfile,"$y = ");
	  PrintString(outputfile,f);
	  fprintf(outputfile,"$ : $");
	  PrintPol(outputfile,*(kld->pol));
	  fprintf(outputfile,"$");
	  if (i%2==1)
	    {
	      fprintf(outputfile,"\\cr");
	    }
	  i++;
	}
if (i%2==1)
	{
	fprintf(outputfile,"\\cr");
	}
fprintf(outputfile,"\n\\endgroup\n");

/* output a table for the Betti numbers and IH betti numbers */

BettiNumbers(w,homology,Ihomology);
h = homology+1;
ih = Ihomology+1;

for (q=0;q<=(p/2);q++)	/* check Poincare duality */
	{
	if (ih[q]!=ih[p-q])
		printf("\nPoincare duality failed for IH betti numbers !!\n\
w = %ld  IH[%d] = %ld  IH[%d] = %ld",
(unsigned long)w,q,(unsigned long)(ih[q]),p-q,(unsigned long)ih[p-q]);
	}

for (q=0;p-q-1>=ROWSIZE;q+=ROWSIZE)	/* print another row */
	{
	fprintf(outputfile,"\n\\medskip\n\n");
	fprintf(outputfile,"\\vbox{\\halign{\
\\tabskip2pt\
\\hfil # &&\\ \\hfil#\\cr\n");	/* preamble */
	fprintf(outputfile,"dimension :");
	for(r=0;r<ROWSIZE;r++)
		{
		fprintf(outputfile,"&%d",q+r);
		}
	fprintf(outputfile,"\\cr\n");
	fprintf(outputfile,"Betti numbers :");
	for(r=0;r<ROWSIZE;r++)
		{
		fprintf(outputfile,"&%ld",(unsigned long)h[q+r]);
		}
	fprintf(outputfile,"\\cr\n");
	fprintf(outputfile,"IH Betti numbers :");
	for(r=0;r<ROWSIZE;r++)
		{
		fprintf(outputfile,"&%ld",(unsigned long)ih[q+r]);
		}
	fprintf(outputfile,"\\cr\n");
	fprintf(outputfile,"}}\n");
	}

fprintf(outputfile,"\n\\medskip\n\n");
fprintf(outputfile,"\\vbox{\\halign{\
\\hfil # &&\\ \\hfil#\\hfil\\cr\n");	/* preamble */
fprintf(outputfile,"dimension :");
for(r=q;r<=p;r++)
	{
	fprintf(outputfile,"&%d",r);
	}
fprintf(outputfile,"\\cr\n");
fprintf(outputfile,"Betti numbers :");
for(r=q;r<=p;r++)
	{
	fprintf(outputfile,"&%ld",(unsigned long)h[r]);
	}
fprintf(outputfile,"\\cr\n");
fprintf(outputfile,"IH Betti numbers :");
for(r=q;r<=p;r++)
	{
	fprintf(outputfile,"&%ld",(unsigned long)ih[r]);
	}
fprintf(outputfile,"\\cr\n");
fprintf(outputfile,"}}\n");

}	/* end of TeXCell */


GetAll()

/***********************************************************************

	Prints the full Kazhdan-Lusztig data for the group W in a TeX-file.
For each element of the Weyl group, the non-trivial Kazhdan-Lusztig 
polynomials are printed (ie the ones that are not obviously equal to one)
for those elements y which are "extremal" with respect to w (ie not
taken up by a generator that takes w down, either on the left or on the 
right).

************************************************************************/

{
char filename[LINESIZE];
FILE *outputfile;
weylnbr_t w;
int a=0;

CALLOC(homology,maxlength+2,weylsize_t,;)
CALLOC(Ihomology,maxlength+2,weylsize_t,;)

printf("Name an output file (default is %c%d.tex)",type,rank);
gets(filename);
if (*filename=='\0')
	{
	sprintf(filename,"%c%d.tex",type,rank);
	}
	
outputfile = fopen(filename,"w");
fprintf(outputfile,"\\hsize16truecm\\vsize23truecm\\nopagenumbers\n");

fprintf(outputfile,"\n\\headline={\\ifnum\\pageno=1\\hfil\\else\n\
\\ifodd\\pageno\\rightheadline\\else\\leftheadline\\fi\\fi}\n\
\\def\\rightheadline{\\tenrm\\runninghead\\hfil\\folio}\n\
\\def\\leftheadline{\\tenrm\\folio\\hfil\\runninghead}\n\
\\def\\runninghead{%c%d}\n",type,rank);

fprintf(outputfile,"\n\\vskip10 mm\n\n\
\\centerline{\\bf Kazhdan-Lusztig polynomials for %c%d}\n\n\
\\vskip10 mm\n",type,rank);

fprintf(outputfile,"This file contains\n");
fprintf(outputfile,"all the Kazhdan-Lusztig\n");
fprintf(outputfile,"polynomials for the Coxeter group\n");
fprintf(outputfile,"$W$ of type %c%d. First of all,\n",type,rank);
fprintf(outputfile,"we choose a standard expression for\n");
fprintf(outputfile,"each element $w$ of the group,\n");
fprintf(outputfile,"which is simply the smallest reduced\n");
fprintf(outputfile,"expression of $w$ in lexicographical\n");
fprintf(outputfile,"order, and we write the elements of $W$\n");
fprintf(outputfile,"in increasing lexicographical\n");
fprintf(outputfile,"order. For each element $w$, we\n");
fprintf(outputfile,"indicate which generators take the length\n");
fprintf(outputfile,"of $w$ down on left (resp.\\ right)\n");
fprintf(outputfile,"multiplication, and we list the\n");
fprintf(outputfile,"elements $y$ which are smaller than\n");
fprintf(outputfile,"$w$ in Bruhat order, and whose length\n");
fprintf(outputfile,"is not taken up by the aforementioned\n");
fprintf(outputfile,"multiplications, with the corresponding\n");
fprintf(outputfile,"Kazhdan-Lusztig polynomial, if it is not 1.\n");
fprintf(outputfile,"Indeed, it is known from the\n");
fprintf(outputfile,"paper of Kazhdan and Lusztig that\n");
fprintf(outputfile,"if $l(sw)<l(w)$ and $l(sy)>l(y)$, then\n");
fprintf(outputfile,"$P_{y,w}=P_{sy,w}$, so it is enough to\n");
fprintf(outputfile,"compute the ones that we list.\n");
fprintf(outputfile,"We also give the ordinary and the\n");
fprintf(outputfile,"intersection homology Betti numbers for\n");
fprintf(outputfile,"the Schubert cell corresponding to $w$ (the\n");
fprintf(outputfile,"dimensions given are the complex\n");
fprintf(outputfile,"dimensions, so they should be multiplied\n");
fprintf(outputfile,"by 2).\n\n");
fprintf(outputfile,"The polynomials marked with a $\\bullet$\n");
fprintf(outputfile,"are the ones for which $\\mu(y,w)\n");
fprintf(outputfile,"\\neq 0$, and $\\mu(y,w)$ is then equal\n");
fprintf(outputfile,"to the leading coefficient in\n");
fprintf(outputfile,"$P_{y,w}$. We obtain thus all the\n");
fprintf(outputfile,"cases for which $\\mu(y,w)\\neq0$,\n");
fprintf(outputfile,"since they are necessarily extremal,\n");
fprintf(outputfile,"except for the ``trivial'' cases\n");
fprintf(outputfile,"where the difference in lengths is 1.\n");

fprintf(outputfile,"\n\\bigskip\n");

for (w=0;w<wsize;w++)
	{
	printf(" [%ld]",(unsigned long)w);
	TeXCell(outputfile,w);
	fprintf(outputfile,"\n\n\\bigskip\n\n");
	}

fprintf(outputfile,"\n\\vfil\\eject\n");
fprintf(outputfile,"\n\
\\leftline{List of all {\\it distinct} Kazhdan-Lusztig polynomials :}\n");
fprintf(outputfile,"\n\\bigskip\n");

PrintTree(outputfile,klstore,&a);

fprintf(outputfile,"\n\n\\bye");
fclose(outputfile);

free(homology);
free(Ihomology);
}	/* end of GetAll */

TestPositive(pol)

/*********************************************************************

	Checks whether the coefficients of pol are positive

**********************************************************************/

	klpol_t pol;

{

int j;
klpol_t a;

for (a=pol+1,j=0;j<= *pol;j++)
	{
	if (*(a++) < 0)
		{
		return(0);
		}
	}
	
return(1);

}	/* end of TestPositive */


PrintStore()

/************************************************************************

	Prints the contents of KLstore

*************************************************************************/

{

FILE *OutputFile;
int a=0;

OutputFile = fopen("KLstore","w");

fprintf(OutputFile,"\\magnification=\\magstep1\n");
fprintf(OutputFile,"\\hsize17truecm\\vsize23truecm\n\n");
fprintf(OutputFile,"\\voffset1truecm\n\n");

fprintf(OutputFile,"Type : %c\n\n",type);
fprintf(OutputFile,"Rank : %d\n\n\\vskip1truecm",rank);
fprintf(OutputFile,"Contents of KLstore :\n\n\\bigskip\n\n");

PrintTree(OutputFile,klstore,&a);

fprintf(OutputFile,"\n\n\\bye");
fclose(OutputFile);

}	/* end of PrintStore */


PrintTree(f,klcell,count)

/***********************************************************************

	Prints the klstore recursively.

************************************************************************/

	FILE *f;
	klcell_t *klcell;
	int *count;

{

if (klcell==0)
	return(0);
PrintTree(f,klcell->left,count);
*count += 1;
fprintf(f,"\\item{%d} :\\quad$",*count);
PrintPol(f,klcell->pol);
fprintf(f,"$\n");
PrintTree(f,klcell->right,count);
return(1);

}	/* end of PrintTree */


PrintPol(f,p)

/**********************************************************************

	Prints the polynomial p in the file f

***********************************************************************/

FILE *f;
klpol_t p;

{
klpol_t p1;
int j;
int warning = 0;

fprintf(f,"1");
if (*p==0)
	{
	return(0);
	}
switch(*(p+2))
	{
	case 0:
		break;
	case 1:
		fprintf(f,"+x");
		break;
	default :
	  if (*(p+2)<0)
	    warning = 1;
		fprintf(f,"+%hdx",*(p+2));
		break;
	}
for (p1=p+3,j=2;j<= *p;j++,p1++)
	{
	  if (*p1<0)
	    warning = 1;
	switch(*p1)
		{
		case 0:
			break;
		case 1:
			fprintf(f,"+x^%d",j);
			break;
		default :
			fprintf(f,"+%hdx^%d",*(p1),j);
			break;
		}
	}

return(warning);

}	/* end of PrintPol */


PrintBPol(f,p)

/*
  Same as PrintPol, for betti-polynomials.
*/

FILE *f;
bettipol_t p;

{
  bettipol_t p1;
  int j;

  fprintf(f,"%ld",*(p+1));
  if (*p==0)
    {
      return(0);
    }
  switch(*(p+2))
    {
    case 0:
      break;
    case 1:
      fprintf(f,"+x");
      break;
    default :
      fprintf(f,"+%ldx",*(p+2));
      break;
    }
  for (p1=p+3,j=2;j<= *p;j++,p1++)
    {
      switch(*p1)
	{
	case 0:
	  break;
	case 1:
	  fprintf(f,"+x^%d",j);
	  break;
	  default :
	  fprintf(f,"+%ldx^%d",*(p1),j);
	  break;
	}
    }

}				/* end of PrintBPol */


SprintPol(s,p)

/**********************************************************************

	Like PrintPol, with output to a string. Returns the number of
characters printed (including null-terminator).

***********************************************************************/

	char *s;
	klpol_t p;

{
klpol_t p1;
int j;
char *s0=s;

*(s++) = '1';
*s = '\0';
if (*p==0)
	{
	return(2);
	}
switch(*(p+2))
	{
	case 0:
		break;
	case 1:
		sprintf(s,"+x");
		break;
	default :
		sprintf(s,"+%hdx",*(p+2));
		break;
	}
for (;*s!=0;s++)
	;
for (p1=p+3,j=2;j<= *p;j++,p1++)
	{
	switch(*p1)
		{
		case 0:
			break;
		case 1:
			sprintf(s,"+x^%d",j);
			break;
		default :
			sprintf(s,"+%hdx^%d",*(p1),j);
			break;
		}
	for (;*s!=0;s++)
		;
	}
	
return(s-s0+1);

}	/* end of SprintPol */


CompletePresentation()

/***********************************************************************

	This function gives a complete presentation of the group W, in the 
form of a set of transformation rules suitable for input into Mathematica.
It will be assumed that the generators are called s[1], ... ,s[rank]. The 
relations given define the group algebra of W; since they form a complete 
presentation, it is guaranteed that their (repeated) application to any 
product of generators will return the normal form. With some more work to 
get distributivity w.r.t. addition, and commutativity w.r.t. scalar 
multiplication, we get the group algebra. Also, adding the relation
s[j]**s[j] -> (q-1) s[j] + q will yield the Hecke algebra of W.

	The idea is that to find the complete presentation, it is enough to 
look at the transition tables constructed by Weyl(), and then for each 
distinguished x in each subquotient to look at the relations of the form 
xs -> tx which are new, in the sense that if x = s_1 ... s_p is the normal 
form, s_2 ... s_p s is also a normal form, but s_1 ... s_p s is not.
Essentially by definition, these relations form the complete presentation.
Actually, this was the "hard case" already met during the construction of 
the automata.

	NOTE :  Mathematica seems rather wasteful in its storage ... As
			written, it will not read the relations for E8 into 1.5 Mb
			of memory, whereas the textfile is < 50 K ...

************************************************************************/

{

FILE *outputfile;
char filename[LINESIZE];
int l;
weylpiece_t *x;
weylarr_t v,vs,w;
weylword_t f;
state_t *state;
generator_t s;

CALLOC(w,rank,weylpiece_t,;)
CALLOC(v,rank,weylpiece_t,;)
CALLOC(vs,rank,weylpiece_t,;)
CALLOC(f,maxlength+1,weylletter_t,;)

sprintf(filename,"%c%d_presentation",type,rank);
outputfile = fopen(filename,"w");

fprintf(outputfile,"%c%d = {\n\
1**x_->x,x_**1->x",type-'A'+'a',rank);	/* get rid of identity elements */

for (l=1,x=w;l<=rank;l++)
	{
	fprintf(outputfile,",\ns[%d]**s[%d]->1",l,l);
	
	for (*x=1,state=table[l-1]+l+1;*x<head[l-1]->quotsize;(*x)++,state+=l+1)
		{
		ArrayToString(w,f);
		StringToArray(f+1,v);	/* cut out first term of w */
		for(s=1;s<=l;s++)
			{
			if(state[s]<0)
				{
				Prod(v,s,vs);
				if (Lastterm(vs)==s)
					{
					fprintf(outputfile,",\n");
					MathPrint(outputfile,f);
					fprintf(outputfile,"**s[%d]",s);
					fprintf(outputfile,"->");
					fprintf(outputfile,"s[%d]**",-state[s]);
					MathPrint(outputfile,f);
					}
				}
			}
		}
	*x++ = 0;
	}

fprintf(outputfile,"}");

fclose(outputfile);
free(v);
free(vs);
free(w);
free(f);

}	/* end of CompletePresentation */


MathPrint(outputfile,f)

/*********************************************************************

	Prints f to outputfile in a format suitable for Mathematica.

**********************************************************************/

	FILE *outputfile;
	weylword_t f;

{
for (;;)
	{
	fprintf(outputfile,"s[%d]",*f);
	if (*++f!='\0')
		{
		fprintf(outputfile,"**");
		}
	else 
		{
		break;
		}
	}
return(0);
}	/* end of MathPrint */


GetMultiplicities()

/*
  Constructs an array with the multiplicities of the Verma module M(w).
  The array can be output to a screen or to a file.
*/

{

  weylnbr_t y,w;
  weylarr_t warray;
  klcoeff_t *mult;
  FILE *outputfile;
  static int screenoutput;
  char filename[LINESIZE];
  char buf[LINESIZE];

  CALLOC(mult,wsize,klcoeff_t,;);
  CALLOC(warray,rank,weylpiece_t,;)

  printf("Enter your element (finish with a carriage-return) :\n");
  GetInput(warray);
  w = LexOrder(warray);

  if (screenoutput==0)
    {
      printf("Name an output file (hit return for stdout):\n");
      gets(filename);
      if (*filename=='\0')
	{
	  outputfile = stdout;
	  screenoutput = 1;
	}
      else
	{
	  outputfile = fopen(filename,"w");
	}
    }
  else
    {
      outputfile = stdout;
    }

  for (y=0;y<wsize;y++)
    {
      mult[y] = Multiplicity(w,y);
    }

  PrintMultiplicities(outputfile,w,mult);

  free(mult);

}    /* end of GetMultiplicities */


PrintMultiplicities(f,w,m)

/*
  Prints the multiplicities for cell no. w in the file outputfile.
*/

    FILE *f;
    weylnbr_t w;
    klcoeff_t *m;

{
  
  weylnbr_t y,y1;
  weylword_t wf;
  lengthcount_t p;
  unsigned lc;
  unsigned maxmult;
  unsigned long total;

  CALLOC(wf, maxlength+1,weylletter_t,;)

  ArrayToString(lexorder+rank*w,wf);
  fprintf(f,"\nw = ");
  PrintString(f,wf);
  fprintf(f,"   (element no. %ld in lexorder)",(unsigned long)w);

  maxmult = 0;
  total = 0;

  for (p=0, y=0; p <= maxlength; p++)
    {
      lc = -1;
      fprintf(f,"\n\nlength %d\n\n#%-5d : ", p, y);
      for (y1=0; y1<betti[p]; y1++,y++)
        {
        if (m[y] > maxmult)
          {
            maxmult = m[y];
          }
        total += m[y];
        if (++lc == MAXLINE)
          {
            fprintf(f,"\n#%-5d : ", y);
            lc = 0;
	  }
        fprintf(f,"%3d ", m[y]);
        }
    }

  fprintf(f,"\n\n");
  fprintf(f,"Largest multiplicity : %d   ",maxmult);
  fprintf(f,"Total : %ld\n\n",total);

  free(wf);

}    /* end of PrintMultiplicities */
