/*
  This is memory.c

  Coxeter version 1.0  Copyright (C) 1991 Fokko du Cloux
  See file main.c for full copyright notice
*/

#include "types.h"
#include "globals.h"

/* 		
		Dynamic memory allocation is usually done in blocks of size 
	BLOCKSIZE, in order to avoid too much overhead from repeated 
	calls to calloc (thisreally seems to make a difference). This 
	approach is used when creatinglists like kl_list, prbruhat and 
	mulist, and for klzone as well (this holds the actual Kazhdan-
	Lusztig polynomials, so that in the programthey can be represented 
	by pointers).
	
		When memory allocation begins, a header of typezoneinfo_t is 
	set up. This contains a pointer to the first and lastblocks, a 
	pointer to the first free element in the current (i.e.last) block, 
	and the free space, in units of the type being stored. Each block
	is assumed to be made of chars, but actually starts with a pointer
	to the next block; hence first and last are char**'s, while firstfree
	is a char*. This makes it easy to free the memory if necessary. 

*/

zoneinfo_t *AllocZone(typesize)

	int typesize;

/**********************************************************************

	Returns a pointer to a new storage zone; the first block is
allocated and initialized to 0; typesize is the size of the type to
be stored in the zone (needed to compute freestore).

***********************************************************************/

{

zoneinfo_t *zone;
static char *ptr;

CALLOC(zone,1,zoneinfo_t,;)

CALLOC(ptr,BLOCKSIZE,char,
	printf(" in allocation of block by AllocZone");)
zone->first = (char **)ptr;
zone->last = zone->first;
*(zone->last) = 0;
zone->firstfree = (char *)zone->first + sizeof(char *);
zone->freestore = (BLOCKSIZE-sizeof(char *))/typesize;

return(zone);

}	/* end of AllocZone */


NewBlock(zone,typesize)

/**********************************************************************

	Adds a new block to the storage zone, to store elements of size
typesize.

***********************************************************************/

	zoneinfo_t *zone;
	int typesize;

{
static unsigned long count;

count++;
CALLOC(*(zone->last),BLOCKSIZE,char,
	printf(" in allocation no. %lu by NewBlock",count);)
zone->last = (char **)*(zone->last);
*(zone->last) = 0;
zone->firstfree = (char *)zone->last + sizeof(char *);
zone->freestore = (BLOCKSIZE-sizeof(char *))/typesize;

return(0);

}	/* end of NewBlock */

FreeZone(zone)

/***********************************************************************

	Frees the storage zone

************************************************************************/

	zoneinfo_t *zone;

{

char *junk;

do
	{
	junk = (char *)zone->first;
	zone->first = (char **)*(zone->first);
	free(junk); 
	}
while (zone->first!=0);

free(zone);

return(0);

}	/* end of FreeZone */


unsigned MemoryCount(n)

/*
  Keeps track of the total memory used
*/
     long n;

{
  static long count;

  count += n;
  if (n>0)
    {
  printf("\nrequest : %d bytes  total : %d",n,count);
}
  if (n<0)
    {
      printf("\nfreed %d bytes  total : %d",-n,count);
    }
  return(count);
}
